/* utf.c - homebrewed functions for dealing with utf8 encoding */
/* Sun Jan 21 11:29:12 EST 2018 -jsj */
/* $Id: utf.c,v 1.25 2021/07/10 19:16:22 malakai Exp $ */

/* Fri Jul  8 10:34:07 AM EDT 2022
 * This code is taken from the Last Outpost Project and made available by the
 * author through the Last Outpost Code Archive: <rahjiii@jeffrika.com>
 * (https://www.last-outpost.com/LO/pubcode) Feel free to use it as you see
 * fit. */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "structs.h"
#include "scnprintf.h"
#include "utils.h"

#include "utf.h"


/* a lot of these translations came indirectly from http://obfuscator.uo1.net/ -jsj */
const struct utf_translation_data utfequiv[] = {
	{'^',"◬↑"},
	{'=',"♨⛁▰ະ▅▆"},
	{'|',"│⚕┤▏▕"},
	{'_',"…▁▂"},
	{'-',"─┉▃▄"},
	{'!',"¡"},
	{'?',"¿"},
	{'/',"╭╯☍☄"},
	{'.',"᛫·"},
	{'(',"⚸"},
	{')',"▏▎"},
	{']',"▋▊"},
	{'}',"▍▌"},
	{'@',"©☠☻☺☹☮⌬⚛"},
	{'$',"¢€¥£"},
	{'*',"☆☏☼★☮♠♣♥♦⚙☘❃"},
	{'\\',"╮╰"},
	{'&',"⚷"},
	{'#',"⊞⍓▇█▉✉"},
	{' ',"        "},
	{'+',"✔✓"},

	{'0',"◒☯ꙮ"},

	{'A',"AÀÁÂÃÄÅĀĂĄǍǞǠȀȂȦΆΑАѦӐӒḀẠẢẤẦẨẬẶἈἉᾈᾉᾸᾹᾺᾼ₳ÅȺẮẰẲẴἌἎἏᾌΆǺẪ♤⚜"},
	{'B',"BƁΒВḂḄḆ"},
	{'C',"CÇĆĈĊČƇʗСҪḈ₢₵ℂⅭϹϾҀ"},
	{'D',"DÐĎĐƉƊḊḌḎḐḒⅮ"},
	{'E',"EÈÉÊËĒĔĖĘĚȄȆȨΕЀЁЕӖḘḚḜẸẺẼẾỀỆḔḖỂỄԐℇƐἙῈЄ"},
	{'F',"FϜḞ₣ҒƑϝғҒ₣"},
	{'G',"GĜĞĠĢƓǤǦǴḠ₲"},
	{'H',"HĤĦȞΗНҢҤӇӉḢḤḦḨḪῌꜦ♔"},
	{'I',"IΊÌÍÎÏĨĪĬĮİƖƗǏȈȊΙΪІЇӀӏḬḮỈỊἸἹῘῙῚǐ1"},
	{'J',"JĴʆЈʃ"},
	{'K',"KĶƘǨΚЌКԞḰḲḴ₭K"},
	{'L',"LĹĻĽĿŁԼḶḸḺḼℒⅬ˪╚"},
	{'M',"MΜМӍḾṀṂⅯ"},
	{'N',"NÑŃŅŇǸΝṄṆṈṊ₦Ɲ∏⋂"},
	{'O',"O0θϑ⍬ÒÓÔÕÖØŌŎŐƆƟƠǑǪǬǾȌȎȪȬȮȰΘΟϴОѲӦӨӪՕỌỎỐỒỔỘỚỜỞỠỢΌΌṌṐṒὈʘṎỖ"},
	{'P',"PƤΡРҎṔṖῬ₱ℙ✍✎"},
	{'Q',"QԚℚ⍜"},
	{'R',"RŔŖŘȐȒṘṚṜṞ℞ɌⱤ⍥⇲"},
	{'S',"SŚŜŞŠȘЅՏṠṢṨṤṦ∫∬"},
	{'T',"TŢŤŦƮȚΤТҬṪṬṮṰ₮ȾΊΊꚌ☥"},
	{'U',"UÙÚÛÜŨŪŬŮŰŲƯǓǕǗǛȔȖԱՍṲṴṶṸỤỦỨỪỬỮỰǙ⊍⊎Մ⊌Ṻ☋"},
	{'V',"VѴѶṼṾ⋁ⅤƲ"},
	{'W',"WŴԜẀẂẄẆẈ₩ƜШ"},
	{'X',"XΧХҲẊẌⅩ⚒"},
	{'Y',"Y¥ÝŶŸƳȲΥΫϓУҮҰẎỲỴỶỸῨῩ"},
	{'Z',"ZŹŻŽƵȤΖẐẒẔ"},
	{'a',"aàáâãäåāăąǎǟǡǻȁȃȧаӑӓḁẚạảấầẩẫậắằẳẵặɑάαἀἁἂἃἄἅἆἇὰάᾀᾁᾂᾃᾄᾅᾆᾇᾰᾱᾲᾳᾴᾶᾷ⍶⍺ɑæ"},
	{'b',"bƀƃƅɒɓḃḅḇþϸƄьҍ"},
	{'c',"cçćĉċčƈςϛсҫḉⅽ¢ϲҁ"},
	{'d',"dďđɖɗḋḍḏḑḓⅾƌժ₫ð"},
	{'e',"eèéêëēĕėęěȅȇȩеѐёҽҿӗḕḗḙḛḝẹẻẽếềểễệεɛϵєϱѳөӫɵ"},
	{'f',"fſḟẛƒғϝ£ƒ"},
	{'g',"gĝğġģǥǧǵɠɡգզցḡɕʛɢ"},
	{'h',"hĥħȟɦɧћիհḣḥḧḩḫẖℏһʜӊ☝"},
	{'i',"iį¡ìíîïĩīĭįıǐȉȋɨɩΐίιϊіїɪḭḯỉịἰἱἲἳὶίῑΐῐῒῖὶ⚵"},
	{'j',"jĵǰȷɟʝјյϳ"},
	{'k',"kķĸƙǩκкҝҟḱḳḵ"},
	{'l',"lŀĺļľłƚǀɫɬɭḷḹḻḽŀ⎩ḹ"},
	{'m',"mɱḿṁṃ₥ⅿ"},
	{'n',"nɴñńņňŉŋƞǹɲɳήηπпբդըղոռրṅṇṉṋἠἡἢἣἤἥἦἧὴήᾐᾑᾒᾓᾔᾕᾖᾗῂῃῄῆῇი☊∩"},
	{'o',"oòóôõöōŏőơǒǫǭȍȏȫȭȯȱʘοόоӧծձօṍṏṑṓọỏốồổỗộớờởỡợὀὁὂὃὄὅὸόσ๐ø"},
	{'p',"pþρрҏթṕṗῤῥ⍴"},
	{'q',"qʠԛգզϙ"},
	{'r',"rŕŗřȑȓɼɽгѓґӷṙṛṝṟгѓґӷ"},
	{'s',"sśŝşšșʂѕԑṡṣṥṧṩß"},
	{'t',"tţťŧƫțʈṫṭṯṱẗȶէե†ԷՒէȽҭ☨"},
	{'u',"uµùúûüũūŭůűųưǔǖǘǚǜȕȗɥμυцկմնսվևṳṵṷṹṻụủứừửữự"},
	{'v',"vʋνѵѷүұṽṿⅴ∨ΰϋύὐὑὒὓὔὕὖὗὺύῠῡῢΰῦῧʋ"},
	{'w',"wŵԝẁẃẅẇẉẘ"},
	{'x',"xϰхҳẋẍⅹ⚔✘✗✖✕"},
	{'y',"yýÿŷƴȳγуўӯӱӳẏẙỳỵỷỹʏ"},
	{'z',"zźżžƶȥʐʑẑẓẕ"},

	{'#',NULL}
};


/* return a single char translation of a utf8 sequence. */
char utf8_to_ascii(char *utfbuf) {

	int i;

	if( ! ((*utfbuf & 0xc0) == 0xc0 )) {
		return(*utfbuf);
	}

	/* this is a slow brute force search. */
	for(i=0;utfequiv[i].utf;i++) {
		if(strstr(utfequiv[i].utf,utfbuf)) {
			return(utfequiv[i].ascii);
		}
	}

	/* this is the last entry in the table. */
	return(utfequiv[i].ascii);

}

/* ascii is gonna be smaller than utf, so its ok to re-write in place. */
char *shrink_to_ascii(char *raw) {
	char *i;
	char *o;
	int len;
	char utfbuf[MAX_UTF_LENGTH];

	for(i=o=raw;*i;i++) {
		len = valid_utf8_sequence(*i,utfbuf);
		if(len >= 1) {
			*o++ = utf8_to_ascii(utfbuf);
		}
	}

	*o = '\0';
	return(raw);
}


/* adds a byte to the utfbuf.  
	if input == '\0', utfbuf is cleared, returns 0
	if input is ascii, *utfbuf is set to input regardless of utfbuf, returns 1
	if input is a UTF start byte, queue is reset regardless of utfbuf, returns -2
	if input is a UTF multibyte byte, and *utfbuf != start, returns -1 and clears utfbuf
	otherwise
	returns the string length of utfbuf if utfbuf is valid.
	returns -1 if the sequence is invalid, sets *utfbuf = '\0'.
	returns -2 if still queueing data
*/
int valid_utf8_sequence(char input, char *utfbuf) {

	int expected_length;
	int len;

	if(input == '\0') {
		/* clear the utfbuf. */
		utfbuf[0] = '\0';
		return(0);
	}

	if ( (input & 0x80) == 0x00 ) {
		/* this is a simple ascii byte. */
		utfbuf[0] = input;
		utfbuf[1] = '\0';
		return(1);
	}

	if( (input & 0xc0) == 0xc0 ) {
		/* this is a utf start byte, start queueing */
		utfbuf[0] = input;
		utfbuf[1] = '\0';
		return(-2);
	}

	if( (input & 0xc0) == 0x80 ) {
		/* this is a utf multi byte */

		/* what total length are we expecting here? */
		if( (*utfbuf & 0xe0) == 0xc0) {
			expected_length = 2;
		} else if( (*utfbuf & 0xf0) == 0xe0) {
			expected_length = 3;
		} else if( (*utfbuf & 0xf8) == 0xf0) {
			expected_length = 4;
		} else if( (*utfbuf & 0xfc) == 0xf8) {
			expected_length = 5;
		} else if( (*utfbuf & 0xfe) == 0xfc) {
			expected_length = 6;
		} else {
			/* uh... why isn't *utfbuf a start byte? */
			goto error;
		}

		/* how much do we have? */
		len = strlen(utfbuf);

		if(len < expected_length) {
			/* append the utf multi to the utfbuf. */
			*(utfbuf+len) = input;
			*(utfbuf+len+1) = '\0';

			if((len+1) < expected_length) {
				/* Looks like we are still queuing. */
				return(-2);
			} else if ((len+1) == expected_length) {
				/* woo hoo, got it! */
				return(expected_length);
			}
		} else {
			goto error;
		}
	}

	error: 
	/* There was some kind of error here. */
	utfbuf[0] = '\0';
	return(-1);

}

/* return the length of the *first* utf sequence in buf */
int utf8_seqlen(char *buf) {
	char *c;
	int len = 0;
	char utfbuf[MAX_UTF_LENGTH];
	
	for(c=buf;*c;c++) {
		if ( (len = valid_utf8_sequence(*c,utfbuf)) > 0) {
			return(len);
		}
	}
	return(0);

}

/* return the number of utf sequences in the utfbuf up to the maxlen (which is
 * buffer bytes, not characters). Which probably isn't the same as the
 * printable length, since I think there are unicode modifier codepoints that
 * don't actually add to the width of the character.  For now though... close
 * enough.*/
int utf8_strnlen(char *buf, size_t maxlen) {
	
	int width = 0;
	char *i,*stop;
	char utfbuf[MAX_UTF_LENGTH];

	/* init */
	valid_utf8_sequence(0,utfbuf);

	i = buf;
	stop = i+maxlen;
	while((*i) && (i<stop) ) {

		/* skip over ansi terminal CSI commands. */
		while(*i && (*i == '\x1B') && (*(i+1) == '[')) {
			for(i=i+2; *i && ( ( (*i >= 0x30) && (*i <= 0x3F) ));i++);
			/* and the final terminating byte. */
			if(*i) {
				i++;
			}
		}

		if(valid_utf8_sequence(*i,utfbuf) > 0) {
			width += 1;
		}
		i++;
	}
	return(width);
}

/* just run to the end of the buffer... */
int utf8_strlen(char *buf) {

	return(utf8_strnlen(buf,(size_t)strlen(buf)));
	
}


